/*
   program : recon

   $Source: /home/yee/C/RCS/backproject.c,v $
   $Author: yee $
   $Date: 1999/03/09 01:52:38 $
   $Revision: 3.8 $

   This file contains the code to implement filtered back projection used in
   image reconstruction.

   This code was written using Kak & Slaney, "Principles of Computerized
   Tomographic Imaging".

   As written, as much as possible, all calculations are done in the frequency
   domain.  
      a) No discrete convolutions are done in the time domain.  Instead, the
         dot products are performed in the frequency domain.  Thus, zero
         padding is needed to avoid interperiod interference artifacts.
      b) At the earliest convenience, the time domain projections are 
         transformed into the frequency domain.  
      c) The backprojection filters are all defined in time domain.  After the
         final filter is obtained, the filter is transformed info frequency
         space.
      d) Only the final backprojection step is done in the time domain.

   Note : The units in frequency space in inverse distance not radians!

   Parameters :
      1) #define RECON_INTERPOLATE for linear interpolation of backprojection
         data or

         #define RECON_NEAREST for nearest neighbor interpolation of data
         (DEFAULT).

      2) #define RECON_DOUBLE|DOUBLE for double precision.  This file defaults
         to #define SINGLE for single precision.

      3) #define RECON_PARALLEL for parallel projection data support.

         #define RECON_PARALLEL_FAN if parallel projection data is to be
         reconstructed as a special case of fan beam data. (slower than native
         parallel projection reconstruction)

         #define RECON_ANGLE_FAN for equiangular fan projection support.

         #define RECON_COLIN_FAN for co-linear fan projection data support.

      4) #define READ_DATA_BY_LINE to reconstruct one projection at a time.

   Assumptions : none

   Bugs : none known

   To Do :
      1) using a smoothing function dramatically changed the reconstructed
         image. need to prove that there isn't a bug in using append mode
         filters. I suspect that there is no bug.
      2) For efficiency, consolidate ALL normalization factors.
         a) inverse FFT
         b) projection filter
         c) final factor in reconstructed array.
         This may be a BAD idea for maintainable code.
      3) Change functions to use an fft that supports real arrays.  This
         should provide a significant performance improvement.
         The projection data should be manipulated as reals.

   $Log: backproject.c,v $
   Revision 3.8  1999/03/09 01:52:38  yee
   *** empty log message ***

   Revision 3.5  1999/01/16 20:49:11  yee
   fixed #define READ_DATA_BY_LINE

   Revision 3.0  1999/01/07 02:13:00  yee
   normalization error fixed.

   Copyright 1997, 1998, 1999 by Conway Yee.  All rights reserved.
 */

#include <math.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "backproject.h"
#include "recon.h"
#include "fft.h"
#include "util.h"
#include "codes.h"
#include "filter.h"
#include "array.h"
#include "mutil.h"

#define FFT_ZP_FACTOR 2      /* zero padding factor (an integer power of 2) */

PRIVATE void DetermineFunctions(CHOICE,BackprojectFunctions *);

PRIVATE void GetProjection(ImageHeader *,int,RECON_COMPLEX *);
PRIVATE void NormalizeReconstructedImage(ImageHeader *,ImageHeader *);
    PUBLIC void normalize(double *,int *,double *);

#ifdef RECON_FAN
  PRIVATE void FanBackprojectProjection(RECON_FLOAT **,int *,double *,
                      RECON_COMPLEX *,int *,double *,double,double,double,
                      double,BackprojectFunctions *);

                             /* find the ray going through the origin that
                                intersects with the row of constant x */
#define DetermineZero(sin_t,cos_t,x) ((fabs(cos_t)<EPSILON)?SIGN(sin_t)*(INT_MAX-1):x*sin_t/cos_t)

#endif RECON_FAN

char backproject_c[] = "$Id: backproject.c,v 3.8 1999/03/09 01:52:38 yee Exp $";
 
/*****************************************************************************/
/*****************************************************************************/

/* this function implements filtered backprojection for both parallel and
   fan beam data.

   As written, all arrays do not use an offset except the final
   reconstructed image and the projection data.  The filters do not
   need to use the offset information.

   All filters are complex arrays. The projection data is also a complex
   array since it must be multiplied by the filter but the final
   reconstructed image is an array of floats. */

PUBLIC void FilteredBackproject
(
 ImageHeader *fh,                  /* header for reconstructed image */
 ImageHeader *ph,                  /* data header */
 double (*window)()                /* windowing filter to be used */
)
{ /* FilteredBackproject() */
  RECON_COMPLEX *H;          /* backprojection filter in freq domain */
  RECON_COMPLEX *q;          /* filtered projection data (without offset) */
  RECON_FLOAT *w=NULL;       /* weighting function for fan data projections */

  double a,b;                /* used in recurrence formulae */
  double cos_t;              /* cos(theta) */
  double sin_t;              /* sin(theta) */

  BackprojectFunctions func; /* finite state machine */

  int zp_ndim=ph->dim[ELE]*FFT_ZP_FACTOR;  /* array dimension used */
  int ndim = ph->dim[ELE];

/************************* Initialize structures */  { /* basic setup */
    { /* allocate data arrays */
      H=(RECON_COMPLEX *)calloc(zp_ndim,sizeof(RECON_COMPLEX));
      q=(RECON_COMPLEX *)calloc(zp_ndim,sizeof(RECON_COMPLEX));

#ifdef RECON_FAN
#ifdef RECON_PARALLEL_NATIVE
            if(ph->type != PARALLEL)
#endif RECON_PARALLEL_NATIVE 
               w=(RECON_FLOAT *)calloc(ndim,sizeof(RECON_FLOAT));
#endif RECON_FAN

      { /* initialize trigonometric recurrence formula */
        double t=ph->tau[PROJ]*ph->lim[LO_PROJ];  /* initial angle, theta */
        InitializeSinCosRecurrence(a,b,cos_t,sin_t,cos(t),sin(t),ph->tau[PROJ]);
      } /* initialize trigonometric recurrence formula */
    } /* allocate data arrays */

    DetermineFunctions(ph->type,&func);  /* set up finite state machine */
  } /* basic setup */

/************************* This block generates the time domain filter.
                           Afterwards, a windowing function is applied.
                           The weighted filter is finally converted into the
                           frequency domain. */

  { /* backprojection filter setup */
                                                          /* generate filter */
    DefineComplexFilter(H,zp_ndim,ndim,BasicBackprojectionFilter,&ph->tau[ELE]);
    AppendComplexFilter(H,zp_ndim,ndim,func.BackprojectionFilterWeight,&ph->tau[ELE]);

    fft(zp_ndim,H,FFT_FORWARD);          /* convert to frequency domain */

    if(window)                                  /* window filter */
      AppendComplexFilter(H,zp_ndim,zp_ndim,window,&ph->tau[ELE]);
  } /* backprojection filter setup */

/************************* For fan beam data, first weight the basic
                           backprojection filter.

                           Then, define filter for projection data. */

#ifdef RECON_FAN
  { /* generate projection weight filter */
    double FloatFree[]={ph->D,ph->tau[ELE]}; /* deg of freedom */

    DefineRealCenteredFilter(w,ndim,ndim,func.ProjectionWeight,FloatFree);
  } /* fan beam data setup */
#endif RECON_FAN

/*************************** Setup is finished. Backproject the data */

  { /* run through projections */
    int i;                                   /* counter */
    double LastProjection=ph->lim[HI_PROJ];  /* last projection */
 
    for(i=ph->lim[LO_PROJ];i<=LastProjection;i++) {
      memset(q,NUL,zp_ndim*sizeof(RECON_COMPLEX));
      GetProjection(ph,i,q);

#ifdef RECON_FAN
#ifdef RECON_PARALLEL_NATIVE
      if(ph->type != PARALLEL)
#endif RECON_PARALLEL_NATIVE
        crrdot(zp_ndim,q,q,w); /* weight projection */
#endif RECON_FAN

      { /* filter projection */
        fft(zp_ndim,q,FFT_FORWARD);
        cdot(zp_ndim,q,q,H);   /* S*H*W,  Note: H already incorporates W */
        fft(zp_ndim,q,FFT_INVERSE);
      } /* filter projection */

      func.BackprojectProjection(fh->data.f,fh->lim,fh->tau,q,ph->lim,ph->tau,
                            i*ph->tau[PROJ],ph->D,cos_t,sin_t,&func);

      UpdateSinCosRecurrence(a,b,sin_t,cos_t);
    }
  } /* run through projections */

  NormalizeReconstructedImage(ph,fh);

/*************************** Cleanup time. Backprojection Done. */

  { /* clean up */
    free(H);
    free(q);
    if(w) free(w);
  } /* clean up */
} /* FilteredBackproject() */

/*****************************************************************************/
/********** utility functions used by all reconstruction algorithms **********/
/*****************************************************************************/

/* this function is used by DefineComplexFilter() to define the
   parallel beam backprojection filter.  Kak & Slaney p.72 eqn 61 */

PUBLIC double BasicBackprojectionFilter
(
 int ndim,           /* dimensionality of data array */
 int i,              /* element number */
 double *tau         /* degrees of freedom, distance between elements */
)
{ /* BasicBackprojectionFilter() */
  double val;  /* returned val */

  if(i==0)                        /* i zero */
    val = 0.25/SQR(*tau);
  else if( i & (int)1)            /* i odd */
    val = -1.0/SQR(i*PI*(*tau));
  else                            /* i even */
    val = 0;

  return (val);
} /* BasicBackprojectionFilter() */

/*****************************************************************************/

/* this function obtains the projection to be processed */

PRIVATE void GetProjection
(
 ImageHeader *ph,   /* image header */
 int i,             /* projection number */
 RECON_COMPLEX *q   /* projection data: 1st element is q[0] */
)
{ /* GetProjection() */
  int ndim=ph->dim[ELE];        /* # elements in a projection */

#ifndef READ_DATA_BY_LINE
  memcpy(q,&(ph->data.c[i][ph->lim[LO_ELE]]),ndim*sizeof(RECON_COMPLEX));
#else
  memset(q,NUL,(FFT_ZP_FACTOR-1)*ndim*sizeof(RECON_COMPLEX));
  mdio(ph->fptr,1,ph->fsize,sizeof(RECON_COMPLEX),q-ph->lim[LO_ELE],fread,&ph->lim[LO_ELE]);
#endif
} /* GetProjection() */

/*****************************************************************************/

/* this function normalizes the reconstructed image */

#ifndef NO_NORMALIZE
PRIVATE void NormalizeReconstructedImage
(
 ImageHeader *ph,  /* Projection Header */
 ImageHeader *fh   /* Reconstructed Image Header */
)
{ /* NormalizeReconstructedImage() */
  double norm;  /* used by the function normalize */

  { /* determine norm */
    double AngleTraversed=ph->dim[PROJ]*ph->tau[PROJ]; /* angle traversed by
                                                         projections */

    switch (ph->type) {
    case PARALLEL:
      norm = PI/ph->dim[PROJ]*                 /* Kak & Slaney p. 67 eqn 45. */
             ph->tau[ELE]*                     /* Kak & Slaney p. 75 eqn 69. */
             TWO_PI/AngleTraversed;            /* Should be ph->tau[ELE] except
                                                  that the projections may not
                                                  go from 0 to 2*PI */
      break;
    case ANGLE:
    case COLIN:
      norm = TWO_PI/ph->dim[PROJ]*             /* Kak & Slaney p. 85 eqn 98 */
             ph->tau[ELE]*                     /* Kak & Slaney p. 97 eqn 97 */
             TWO_PI/AngleTraversed;            /* Should be ph->tau[ELE] except
                                                  that the projections may not
                                                  go from 0 to 2*PI */

      break;
    default:
      exit;
    }
  } /* determine norm */
                                                          /* normalize image */
  mdaccess(fh->ndim,fh->fsize,fh->data.f,normalize,fh->lim,NULL,&norm); 
} /* NormalizeReconstructedImage() */
#endif NO_NORMALIZE

/*****************************************************************************/

/* this function is called by mdaccess() in NormalizeReconstructedImage()
   to normalize the image array */

PUBLIC void normalize
(
 double *num,       /* array element */
 int *IntFree,      /* integer degrees of freedom */
 double *FloatFree  /* floating point degrees of freedom */
)
{ /* normalize() */
 *num *= *FloatFree; /* FloatFree[] = norm, the normalization factor */
} /* normalize() */

/*****************************************************************************/

/* this function performs back projection on all fan beam data.

   NOTE: cos(theta) and sin(theta) are sent in as parameters to allow the
         use of recurrence formulae.   */

#ifdef RECON_FAN
PRIVATE void FanBackprojectProjection
(
 RECON_FLOAT **f,           /* reconstructed image */
 int *flim,                 /* dimensionality of f[][] */
 double *ftau,              /* dist between samples in reconstructed image */
 RECON_COMPLEX *q,          /* projection data */
 int *qlim,                 /* dimensionality of q[] */
 double *qtau,              /* distance between samples in projection data */
 double theta,              /* angle of projection data */
 double D,                  /* dist between focal pt and center of rotation */
 double cos_t,              /* cos(angle of projection data) */
 double sin_t,              /* sin(angle of projection data) */
 BackprojectFunctions *func /* ptr to structure containing function choices */
)
{ /* FanBackprojectProjection() */
  int i;               /* counter */

  q -= qlim[LO_ELE];  /* FanBackprojectProjection() uses offset information */

  for(i=flim[LOX];i<=flim[HIX];i++) { /* loop on i */
    int dir;                                     /* direction to increment j */
    double x=i*ftau[X];                                 /* X position */
    int start=(int)floor(DetermineZero(sin_t,cos_t,x)); /* limits on looping */

    for(dir= -1;dir<=1;dir+=2) { /* loop on direction about zero */
      int j;                       /* counter */
      double ProjectionPosition;   /* projected float position of (i,j) */
      BOOLEAN out=FALSE;           /* outside of projection limits */

                                    /* kludge: too many comparisions */
      j=((dir== -1)?MIN(start,flim[HIY]):MAX(start+1,flim[LOY]));
      for(;(j>=flim[LOY])&&(j<=flim[HIY])&&(!out);j+=dir) {  /*loop on j*/
        double y=j*ftau[Y];              /* Y position */

        ProjectionPosition
            =func->BackprojectPoint(x,y,qtau,ftau,sin_t,cos_t,D)/qtau[ELE];

#ifdef RECON_INTERPOLATE
        { /* interpolate */
          int lo=(int)floor(ProjectionPosition); /* low end of interpolation */
          double fr=ProjectionPosition-lo;       /* fraction between pixels */
         
          if((lo>=qlim[LO_ELE])&&(lo<qlim[HI_ELE]))
            f[i][j]+=(fr*q[lo].r+(1-fr)*q[lo+1].r)/
                        func->BackprojectNormalize(D,x,y,sin_t,cos_t);
          else
            out=TRUE;
        } /* interpolate */
#else RECON_NEAREST
        { /* nearest neighbor */
          int ele=NINT(ProjectionPosition);       /* element of q used */

          if((ele>=qlim[LO_ELE])&&(ele<=qlim[HI_ELE]))
            f[i][j]+=q[ele].r/func->BackprojectNormalize(D,x,y,sin_t,cos_t);
          else
            out=TRUE;
        } /* nearest neighbor */
#endif
      } /* loop on j */
    } /* loop on direction about zero */
  } /* loop on i */
} /* FanBackprojectProjection() */
#endif RECON_FAN

/*****************************************************************************/

/* Setup finite state machine. In other words, determine the proper
   backprojection functions for fan beam data */

#ifdef RECON_FAN
PRIVATE void DetermineFunctions
(
 CHOICE type,                /* data type */
 BackprojectFunctions *f     /* ptr to struct containing function choices */
)
{ /* DetermineFunctions() */
  switch(type) {
#ifdef RECON_PARALLEL
  case PARALLEL:
#ifndef RECON_PARALLEL_FAN
    f->BackprojectionFilterWeight = NULL;
    f->ProjectionWeight           = NULL;
    f->BackprojectProjection      = ParallelBackprojectProjection;
    f->BackprojectPoint           = NULL;
#else RECON_PARALLEL_FAN
    f->BackprojectionFilterWeight = ParallelBackprojectionFilterWeight;
    f->ProjectionWeight           = ParallelProjectionWeight;
    f->BackprojectProjection      = FanBackprojectProjection;
    f->BackprojectPoint           = ParallelBackprojectPoint;
    f->BackprojectNormalize       = P2;
#endif RECON_PARALLEL_FAN
    break;
#endif RECON_PARALLEL
#ifdef RECON_ANGLE_FAN
  case ANGLE:
    f->BackprojectionFilterWeight = AngleBackprojectionFilterWeight;
    f->ProjectionWeight           = AngleProjectionWeight;
    f->BackprojectProjection      = FanBackprojectProjection;
    f->BackprojectPoint           = AngleBackprojectPoint;
    f->BackprojectNormalize       = L2;
    break;
#endif RECON_ANGLE_FAN
#ifdef RECON_COLIN_FAN
  case COLIN:
    f->BackprojectionFilterWeight = ColinBackprojectionFilterWeight;
    f->ProjectionWeight           = ColinProjectionWeight;
    f->BackprojectProjection      = FanBackprojectProjection;
    f->BackprojectPoint           = ColinBackprojectPoint;
    f->BackprojectNormalize       = U2;
    break;
#endif RECON_COLIN_FAN
  default:
    perror("DetermineFunctions : unsupported data type ");
  }
} /* DetermineFunctions() */
#endif RECON_FAN
